c
c  program to resample a waveform file
c
c  jh, July 2024
c

      implicit none
      include 'seidim.inc'                    ! for giving dimensions
      include 'libsei.inc'                    ! for all seisan subroutines
      include 'waveform.inc'                  ! waveform data
      include 'seisan.inc'                    ! seisan general parameters
      integer nstat,nphase,nhead,nrecord,id   ! for reading s-file
      character*1 type,exp                    ! -----------------
      character*80 sfilname                   ! s-file name
      character*80 data(max_data)             ! one s-file
      character*80 infile                     ! input file name
      character*80 outfile                    ! output file name
      integer npresent                        ! number of wav files present
      character*80 text                       ! general text string
      character*80 question                   ! question
      integer save_ncomp                      ! number of different components
      integer appendFlag                      ! append flag for mseed write
      integer compFlag                        ! compression flag, 0: steim1, 1 steim2
      integer recSize                         ! record size to write mseed
      character*3 seed_comp                   ! seed component
      integer seiclen
      integer nars                            ! number of arguments
      character*80 arg(100)                   ! arguments

c
c   for resampling
c
      integer skip                            ! skipping for resampling
      integer passes                          ! number of passes through filter
      integer npoles                          ! number of poles
      real dc                                 ! dc
      real filter                             ! antialias filter     

      integer k,i,in,isamp,ifile
c
c  use a waveform file name directly or the waveform files in an s-file
c

      write(6,*)
     *'Program to resample one or several waveform files.'
      write(6,*)
     *'The input format can be any of the SEISAN accepted formats.'
      write(6,*)
     *'The output format is only MSEED.'
      write(6,*)
     *'Input files can be one file, files in filenr.lis or an S-file'
      write(6,*)
     *'The resampling is by a decimation factor e.g. 5'
      write(6,*)
     *'All channels in all files will be resampled the same way'
      write(6,*)
     *'The antialias filter is defualt an '//
     *'8 pole low pass Butterworth filter.'
      write(6,*)
     *'By default the filter only passes foreward (passes=1)'
      write(6,*)
     *'It is up up the user to select an appropriate filter.'
      write(6,*)
     *'All parameters can also be selected using flags:'
      write(6,*) '-file   : file name of one waveform file'
      write(6,*) '-skip   : the decimation factor'
      write(6,*) '-filter : frequency of the LP filter'
      write(6,*) '-poles  : number of poles of the filter'
      write(6,*) '-passes : if given, filter will pass both ways'
      write(6,*)
     *'The output file names are the input file names with'//
     *' _resamp appended'

       write(6,*)
      

c
c  initialize 
c
      call wav_init
      call wav_mem_init
      call get_seisan_def

c
c   get arguments
c
      call get_arguments(nars,arg)


c  Compression flag   0 - STEIM1     1 - STEIM2
c 
      compFlag = 0
c
c  Record Size
c
      recSize = 4096
c
c   check arguments
c
      infile=' '
      filter=0.0
      npoles=0
      passes=0
      skip=0

      if(nars.gt.0) then
         do i=1,nars
          if(arg(i)(1:5).eq.'-file') infile=arg(i+1)
          if(arg(i)(1:7).eq.'-filter') read(arg(i+1),*) filter
          if(arg(i)(1:6).eq.'-poles') read(arg(i+1),*) npoles
          if(arg(i)(1:7).eq.'-passes') passes=2
          if(arg(i)(1:5).eq.'-skip') read(arg(i+1),*) skip 
        enddo
      endif

      if(npoles.eq.0) npoles=8
      if(passes.eq.0) passes=1

      if(skip.eq.0) then
          write(6,*)'Give decimation '
          read(5,*) skip
      endif
      if(filter.eq.0.0) then
         write(6,*)'Give low pass filter'
         read(5,*) filter
      endif
      if(infile.ne.' ') then
         in=0                    ! only single wav files
         wav_filename(1)=infile
         wav_nfiles=1
         goto 1001
      endif
c
c   get s-file name or wav name
c
      write(6,*)' S-file name, return for wav-name directly'
      read(5,'(a)') sfilname
      if(sfilname.ne.' ') then
c
c   open s-file
c
         open(1,file=sfilname,status='old')
c
c  read s-file
c
         call indata
     *   (1,nstat,nphase,nhead,nrecord,type,exp,data,id)
         close(1)

c
c   get waveform file names, could be several
c
         call auto_tr(data,nhead,nrecord,wav_nfiles,wav_filename)

      else
c
c   get file name directly, can be one or many
c
        in=0
        question=' File name, # or filenr.lis for all'
        call filename(question,wav_filename(1))
        if(wav_filename(1)(1:10).eq.'filenr.lis'.or.
     *  wav_filename(1)(1:10).eq.'FILENR.LIS') then
           open(8,file='filenr.lis',status='old',err=20)
           goto 21
 20        continue
           write(6,*)' No filenr.lis'
           stop
 21        continue
           in=1
        endif
        wav_nfiles=1     ! there is only one file at the time in this case
      endif
c
c   file loop if many files
c
 1000 continue
      if(in.eq.1) then
         read(8,'(7x,a)')wav_filename(1) 
         if(wav_filename(1)(1:4).eq.'    ') goto 2000 ! the end
      endif
c
c   find how many waveform files are present,
c   replace the origial names without path, with the present files
c   full path like
c
       npresent=0                      ! no files initially
       do i=1,wav_nfiles
          call  get_full_wav_name(wav_filename(i),text)            
          if(text.ne.' ') then
             npresent=npresent+1
             wav_filename(npresent)=text
          endif
      enddo
c
c   print how many files were found
c
      wav_nfiles=npresent
      write(6,*)' Number of wav-files present', wav_nfiles

c
c   terminate if no waveform files
c
      if(wav_nfiles.eq.0) stop

c
      wav_nchan=0

 1001 continue   ! get her if only one file from prompt     
c
c   loop to read all headers of all files
c
      do ifile=1,wav_nfiles
      write(6,'(1x,a)') wav_filename(ifile)
c
c   read all headers of file i
c
      call read_wav_header(ifile)
c
c   output possible errors
c
         if(wav_error_message.ne.' ') write(6,'(1x,a)') 
     *   wav_error_message
c
c   write out the format
c
         write(6,'(1x,a,a)') 'Format ',wav_file_format(ifile)

cxx      enddo
c
      write(6,*)' Total number of channels available:',wav_nchan
c
c   terminate if no channels
c
      if(wav_nchan.eq.0) stop
c
c   write some input info for each channnel on screen
c
      write(6,'(a,a)')
     *'CHA STAT  COMP  YEAR MO DA HR MI    SEC   NSAMP ',
     *'    RATE    DELAY DURATION'
      do i=1,wav_nchan
         write(6,'(i3,1x,a,1x,a,1x,i5,4i3,f7.3,i8,3f9.3)')
     *   i, wav_stat(i),wav_comp(i),
     *   wav_year(i),wav_month(i),wav_day(i),wav_hour(i),wav_min(i),
     *   wav_sec(i),wav_nsamp(i),wav_rate(i),wav_delay(i),
     *   wav_duration(i)
      enddo
 
      outfile=wav_filename
     *(ifile)(1:seiclen(wav_filename(ifile)))//'_resamp'
c
c
c----------------------------------------------------------
c  resample, one channel at a time
c-----------------------------------------------------------
c
      do k=1,wav_nchan
c
c   read whole channel
c
         call wav_read_channel(k)
c
c   write first 10 samples on screen, remember originaly they are integer
c   however, when read by SEISAN, they are converted to real
c
         write(6,*)
         write(6,*)' 8 first samples from channel ',k
         write(6,'(8i10)') (int(signal1(i)),i=1,8)   ! signal1 is a common block variable 
c
c   resample if filter ok
c

         if(wav_rate(k).le.filter*2.0) then
            write(6,*)
     *      'Sample rate too low for filter, skip filter and decimation'
            goto 500
         endif
         
c
c   remove dc and filter
c


         dc=0.0
         do i=1,wav_nsamp(k)
            dc=dc+signal1(i)
         enddo
         dc=dc/wav_nsamp(k)
        

         do i=1,wav_nsamp(k)
            signal1(i)=signal1(i)-dc
         enddo
c
c   filter
c
         call recfil(signal1,wav_nsamp(k),signal1,'BU      ',0.0,0.0,
     *   npoles,'LP      ',0.0,filter,1.0/wav_rate(k),passes)
         write(6,*)' After filtering'
         write(6,'(8i10)') (int(signal1(i)),i=1,8)    
c
c   resample
c
         write(6,'(a,f7.2,i10)')
     *   'Input sample rate and number of samples:  ',
     *    wav_rate(k),wav_nsamp(k)
         isamp=0
         do i=1,wav_nsamp(k),skip
            isamp=isamp+1
            signal_int(isamp)=signal1(i)
         enddo
         wav_nsamp(k)=isamp
         wav_rate(k)=wav_rate(k)/skip
         write(6,'(a,f7.2,i10)')
     *   'Output sample rate and number of samples: ',wav_rate(k),isamp
c
c   write channel in mseed format
c
 500     continue
c
c  Append flag        0 - Replace    1 - Append
c
         if(k.eq.1) then
            appendFlag = 0
         else
            appendFlag = 1
         endif 

         seed_comp(1:2)=wav_comp(k)(1:2)
         seed_comp(3:3)=wav_comp(k)(4:4)

         call seed_channel_write(outfile,
     *   wav_stat(k), seed_comp, wav_network(k), wav_location(k),
     *   wav_year(k),wav_month(k),wav_day(k),wav_hour(k),wav_min(k),
     *   wav_sec(k),wav_rate(k),
     *   wav_nsamp(k),signal_int,
     *   appendFlag,compFlag,recSize,wav_time_error(k))
 

      enddo
c
c   if a list of files, go back to read next
c
      if(in.eq.1) goto 1000
      enddo                    ! end of multiple files in an S-file

c
c   the end, no more files
c
 2000 continue
  
c
      write(6,*)
      write(6,'(a)')     'Parameters used'
      Write(6,'(a,i8)')  'Skip    :',skip
      Write(6,'(a,f8.2)')'Filter  :',filter
      Write(6,'(a,i8)')  'Poles   :',npoles
      Write(6,'(a,i8)')  'Passes  :',passes
      write(6,*)

      write(6,'(a)') 
     *'Output in file(s) is input file appended with _resamp'

      stop
      end

